# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip import exceptions
from f5.bigip.rest_collection import icontrol_folder
from f5.bigip.rest_collection import icontrol_rest_folder
from f5.bigip.rest_collection import log
from f5.common import constants as const
from f5.common.logger import Log

import datetime
import json
import os
import re
import time
import urllib2

from OpenSSL import crypto


class SSL(object):
    """Interface for SSL related REST methods """

    OBJ_PREFIX = 'uuid_'

    def __init__(self, ltm_instance):
        self.bigip = ltm_instance.bigip

        self.bigip.icontrol.add_interfaces(['Management.KeyCertificate',
                                            'LocalLB.ProfileClientSSL'])
        self.mgmt_keycert = self.bigip.icontrol.Management.KeyCertificate
        self.lb_clientssl = self.bigip.icontrol.LocalLB.ProfileClientSSL

    class Certificate(object):
        """Provides import and download capability for X.509 certificates.

        """

        __key_passphrase__ = None
        __verified__ = False

        key_length = 0
        key_passphrase_required = False
        subject_cn = None
        serial_number = None
        certificate_data = None
        expiration_date = None
        issuer_cn = None
        version = None

        certifcate_id = None

        def __init__(self,
                     name=None,
                     cert=None,
                     key=None,
                     pkcs12=None,
                     passphrase=None):
            if name:
                self.certifcate_id = name
            if cert:
                try:
                    self.get_PEM_certificate(url=cert)
                except ValueError:
                    self.certificate_from_PEM_data(PEM_data=cert)
            if key:
                try:
                    self.get_PEM_key(url=key, key_passphrase=passphrase)
                except ValueError:
                    self.key_from_PEM_data(PEM_data=cert)

        def id_from_subject_cn(self):
            if self.subject_cn:
                self.certificate_id = \
                    str(self.subject_cn).replace("*", 'wildcard') + \
                    "-" + \
                    re.sub('[^A-Za-z0-9]+', '', str(self.serial_number)) + \
                    "-"+str(self.expiration_date.isoformat())

        def certificate_from_PEM_data(self, PEM_data=None):
            if PEM_data:
                x509cert = crypto.load_certificate(  # @UndefinedVariable
                    crypto.FILETYPE_PEM,  # @UndefinedVariable
                    PEM_data
                )
                tm = list(
                    time.strptime(
                        x509cert.get_notAfter()[:8], "%Y%m%d"
                    )
                )[:6]
                tm.append(0)
                tm.append(None)

                self.subject_cn = x509cert.get_subject().__getattribute__('CN')
                self.serial_number = x509cert.get_serial_number()

                self.certificate_data = PEM_data
                self.expiration_date = datetime(*tm).date()
                self.issuer_cn = x509cert.get_issuer().__getattribute__('CN')
                self.version = int(x509cert.get_version())

                if not self.certificate_id:
                    self.id_from_subject_cn()

            else:
                raise ValueError("PEM data for certificate must not be None")

        def key_from_PEM_data(self,
                              PEM_data=None,
                              key_passphrase=""):
            try:

                x509key = crypto.load_privatekey(  # @UndefinedVariable
                    crypto.FILETYPE_PEM,  # @UndefinedVariable
                    PEM_data,
                    key_passphrase
                )
                self.key_data = PEM_data
                self.bit_length = int(x509key.bits())

                if key_passphrase != "":
                    self.__key_passphrase__ = key_passphrase
                    self.key_passphrase_required = True
                    self.key_passphrase = key_passphrase

            except crypto.Error as error:  # @UndefinedVariable
                if "bad decrypt" in error.message[0]:
                    raise ValueError("key passphrase was incorrect")
                else:
                    raise

        def from_PKCS12_data(self, pkcs12_data=None, import_passphrase=None):
            if len(pkcs12_data):
                if not import_passphrase:
                    raise Exception("PKCS12 requires a import password.")

                try:
                    pkcspackage = crypto.load_pkcs12(  # @UndefinedVariable
                        pkcs12_data,
                        import_passphrase
                    )
                except crypto.Error as error:  # @UndefinedVariable
                    if "mac verify failure" in error.message[0]:
                        raise Exception(
                            "import passphrase for PKCS12 was incorrect"
                        )
                    else:
                        raise
                x509cert = pkcspackage.get_certificate()
                tm = list(
                    time.strptime(
                        x509cert.get_notAfter()[:8], "%Y%m%d"
                    )
                )[:6]
                tm.append(0)
                tm.append(None)

                self.subject_cn = \
                    x509cert.get_subject().__getattribute__('CN')
                self.serial_number = int(x509cert.get_serial_number())
                self.certificate_data = \
                    crypto.dump_certificate(  # @UndefinedVariable
                        crypto.FILETYPE_PEM,  # @UndefinedVariable
                        x509cert
                    )
                self.expiration_date = datetime(*tm).date()
                self.issuer_cn = \
                    x509cert.get_issuer().__getattribute__('CN')
                self.version = int(x509cert.get_version())

                private_key = pkcspackage.get_privatekey()
                self.key_data = crypto.dump_privatekey(  # @UndefinedVariable
                    crypto.FILETYPE_PEM,  # @UndefinedVariable
                    private_key
                )
                self.bit_length = int(private_key.bits())
                self.key_passphrase_required = False
                self.key_passphrase = None

                if not self.certificate_id:
                    self.id_from_subject_cn()

            else:
                raise Exception("PKCS12 data must not be None")

        def get_PEM_certificate(self, url=None):
            if url:
                reader = urllib2.urlopen(url)
                certificate = reader.read()
                if len(certificate):
                    self.certificate_from_PEM_data(PEM_data=certificate)
                else:
                    raise ValueError(
                        "PEM certificate URL %s read with no content."
                        % url
                    )
            else:
                raise ValueError(
                    "Must supply a URL to download PEM certificate." +
                    " i.e. file:///path/host.crt"
                )

        def get_PEM_key(self, url=None, key_passphrase=""):
            if url:
                reader = urllib2.urlopen(url)
                try:
                    key = reader.read()
                    if len(key):
                        self.key_from_PEM_data(PEM_data=key,
                                               key_passphrase=key_passphrase)
                except crypto.Error as error:  # @UndefinedVariable
                    if "bad decrypt" in error.message[0]:
                        raise ValueError("key passphrase was incorrect")
                    else:
                        raise
            else:
                raise ValueError("Must supply a URL to download PEM key." +
                                 " i.e. file:///path/host.crt")

        def get_PKCS12(self, url=None, import_passphrase=None):
            if url:

                if not import_passphrase:
                    raise ValueError("PKCS12 requires a import password.")
                reader = urllib2.urlopen(url)
                pkcspackage = reader.read()

                if len(pkcspackage):
                    self.from_PKCS12_data(pkcs12_data=pkcspackage,
                                          import_passphrase=import_passphrase)
                else:
                    raise ValueError(
                        "PKCS12 URL %s read with no content." % url
                    )
            else:
                raise ValueError(
                    "Must supply a URL to download PEM certificate." +
                    " i.e. file:///path/host.crt"
                )

    @log
    @icontrol_folder
    def create_clientssl_profile_for_certificate(
            self,
            certificate=None,
            parent_profile='/Common/clientssl',
            folder='Common'):

        """Creates tenant ssl profile for the specified certificate folder

        to create the ssl client profile
        """
        if not isinstance(certificate, SSL.Certificate):  # @UndefinedVariable
            raise Exception('certificate is not an instance of Certificate')

        profile_name = certificate.certificate_id
        user_default_parent = True

        if not parent_profile == '/Common/clientssl':
            parent_profile_name = os.path.basename(parent_profile)
            parent_profile_folder = os.path.dirname(parent_profile)
            if not self.client_profile_exists(name=parent_profile_name,
                                              folder=parent_profile_folder):
                raise ValueError('parent clientssl profile %s does not exist'
                                 % parent_profile)
            user_default_parent = False

        if not self.client_profile_exits(name=profile_name, folder=folder):
            # add certificates to group
            self.mgmt_keycert.certificate_import_from_pem(
                mode='MANAGEMENT_MODE_DEFAULT',
                cert_ids=[profile_name],
                pem_data=[certificate.certificate_data],
                overwrite=True
            )
            self.mgmt_keycert.key_import_from_pem(
                mode='MANAGEMENT_MODE_DEFAULT',
                key_ids=[profile_name],
                pem_data=[certificate.key_data],
                overwrite=True
            )
            # add SSL profile
            profile_string_cert = \
                self.lb_clientssl.typefactory.create('LocalLB.ProfileString')
            profile_string_cert.value = profile_name+".crt"
            profile_string_cert.default_flag = False
            profile_string_key = \
                self.lb_clientssl.typefactory.create('LocalLB.ProfileString')
            profile_string_key.value = profile_name+".key"
            profile_string_key.default_flag = False
            self.lb_clientssl.create_v2(
                profile_names=[profile_name],
                keys=[profile_string_key],
                certs=[profile_string_cert]
            )

            if not user_default_parent:
                profile_string_defaults = \
                    self.lb_clientssl.typefactory.create(
                        'LocalLB.ProfileString'
                    )
                profile_string_defaults.value = parent_profile
                profile_string_defaults.default_flag = False
                self.lb_clientssl.set_default_profile(
                    profile_names=[profile_name],
                    defaults=[profile_string_defaults]
                )
            if certificate.__key_passphrase__:
                profile_string_passphrase = \
                    self.lb_clientssl.typefactory.create(
                        'LocalLB.ProfileString'
                    )
                profile_string_passphrase.value = \
                    certificate.__key_passphrase__
                profile_string_passphrase.default_flag = False
                self.lb_clientssl.set_passphrease(
                    profile_names=[profile_name],
                    passphrases=[profile_string_passphrase]
                )

    @log
    @icontrol_folder
    def remove_clientssl_profile_and_certificate(self,
                                                 certificate=None,
                                                 folder='Common'):
        """Removes a client ssl profile.

        """
        if not isinstance(certificate, SSL.Certificate):  # @UndefinedVariable
            raise Exception('certificate is not an instance of Certificate')

        profile_name = certificate.certifcate_id

        if self.client_profile_exits(name=profile_name, folder=folder):
            # remove ssl profile
            self.lb_clientssl.delete_profile([profile_name])
            # remove certificate
            self.mgmt_keycert.certificate_delete(
                mode='MANAGEMENT_MODE_DEFAULT',
                cert_ids=[profile_name]
            )
            # remove key
            self.mgmt_keycert.key_delete(
                mode='MANAGEMENT_MODE_DEFAULT',
                key_ids=[profile_name]
            )

    @log
    @icontrol_rest_folder
    def all_client_profile_names(self, name=None, folder='Common'):
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/ltm/profile/client-ssl/'
        request_url += '?$select=name,partition'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT
        )
        profile_names = []
        if response.status_code < 400:
            return_obj = json.loads(response.text)
            if 'items' in return_obj:
                for profile in return_obj['items']:
                    print("%s" % profile)
                    profile_name = '/' + \
                                   profile['partition'] + \
                                   '/' + \
                                   profile['name']
                    profile_names.append(profile_name)
        elif response.status_code != 404:
            Log.error('pool', response.text)
            raise exceptions.PoolQueryException(response.text)
        return profile_names

    @log
    @icontrol_rest_folder
    def client_profile_exits(self, name=None, folder='Common'):
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/ltm/profile/client-ssl/'
        request_url += '~' + folder + '~' + name
        request_url += '?$select=name'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        elif response.status_code != 404:
            Log.error('pool', response.text)
            raise exceptions.PoolQueryException(response.text)
        return False



# Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import Collection
from f5.bigip.resource import MissingRequiredCreationParameter
from f5.bigip.resource import Resource


class RequireOneOf(MissingRequiredCreationParameter):
    pass


class SNATCollection(Collection):
    def __init__(self, ltm):
        super(SNATCollection, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [SNAT]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:snat:snatstate': SNAT}


class SNAT(Resource):
    def __init__(self, snat_collection):
        '''This represents a SNAT.

        "origins" is our first example of a dict attribute, it appears to
        behave as expected.
        '''
        super(SNAT, self).__init__(snat_collection)
        self._meta_data['required_json_kind'] = 'tm:ltm:snat:snatstate'
        self._meta_data['required_creation_parameters'].update(
            ('partition', 'origins'))

    def create(self, **kwargs):
        rcp = self._meta_data['required_creation_parameters']
        required_singles = set(('automap', 'snatpool', 'translation'))
        pre_req_len = len(kwargs.keys())
        if len(rcp - required_singles) != (pre_req_len-1):
            error_message = 'Creation requires one of the provided k,v:\n'
            for req_sing in required_singles:
                try:
                    req_val = kwargs.pop(req_sing)
                except KeyError:
                    req_val = ''
                error_message = error_message + str(req_sing) + ', ' +\
                    str(req_val) + '\n'
            raise RequireOneOf(error_message)
        self._create(**kwargs)
        return self



# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip import exceptions
from f5.bigip.rest_collection import log
from f5.common import constants as const

import json
import re
import time


class Stat(object):
    """Class for accessing bigip statistics """
    def __init__(self, bigip):
        self.bigip = bigip

    @log
    def get_global_statistics(self):
        """Get platform global statisitcs """
        request_url = self.bigip.icr_url + '/sys/performance/all-stats'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)

        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'apiRawValues' in response_obj:
                sr = {
                    'Sys::Performance System': {
                        'System CPU Usage': {
                            'Utilization': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            }
                        },
                        'Memory Used': {
                            'TMM Memory Used': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            },
                            'Other Memory Used': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            },
                            'Swap Memory Used': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            }
                        }
                    },
                    'Sys::Performance Connections': {
                        'Active Connections': {
                            'Connections': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            }
                        },
                        'Total New Connections': {
                            'Client Connections': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            },
                            'Server Connections': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            }
                        },
                        'HTTP Requests': {
                            'HTTP Requests': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            }
                        }
                    },
                    'Sys::Performance Throughput': {
                        'Throughput(bits)': {
                            'In': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            },
                            'Out': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            }
                        },
                        'SSL Transactions': {
                            'SSL TPS': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            }
                        },
                        'Throughput(packets)': {
                            'In': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            },
                            'Out': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            }
                        }
                    },
                    'Sys::Performance Ramcache': {
                        'RAM Cache Utilization': {
                            'Hit Rate': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            },
                            'Byte Rate': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            },
                            'Eviction Rate': {
                                'current': 0,
                                'average': 0,
                                'max': 0
                            }
                        }
                    }
                }
                stats_display = response_obj['apiRawValues']['apiAnonymous']
                lines = str(stats_display).split('\n')
                sec = None
                div = None
                since = None
                for line in lines:
                    if len(line) > 2:
                        for this_section in sr.keys():
                            if str(line).startswith(this_section):
                                if sec:
                                    if not (sec == this_section):
                                        sec = this_section
                                        div = None
                                else:
                                    sec = this_section
                        if sec:
                            for division in sr[sec].keys():
                                if str(line).startswith(division):
                                    try:
                                        since_idx = line.index('since')
                                        end_since_idx = line.index(')',
                                                                   since_idx)
                                        since = line[since_idx+6:end_since_idx]
                                    except ValueError:
                                        pass
                                    div = division
                        if div:
                            for fields in sr[sec][div].keys():
                                for field in fields:
                                    if str(line).startswith(field):
                                        values = re.split(r'\s{2,}', line)
                                        if len(values) == 4:
                                            if values[0] in fields:
                                                sr[sec][div][values[0]] = \
                                                    {
                                                        'current': values[1],
                                                        'average': values[2],
                                                        'max': values[3]
                                                    }
                sr['since'] = since
                return sr
            return None
        elif response.status_code == 404:
            return None
        else:
            raise exceptions.SystemQueryException(response.text)

    @log
    def get_composite_score(self):
        """Get composite score """
        gs = self.get_global_statistics()
        cpu_score = self.get_cpu_health_score(gs) * \
            const.DEVICE_HEALTH_SCORE_CPU_WEIGHT
        mem_score = self.get_mem_health_score(gs) * \
            const.DEVICE_HEALTH_SCORE_MEM_WEIGHT
        cps_score = self.get_cps_health_score(gs) * \
            const.DEVICE_HEALTH_SCORE_CPS_WEIGHT

        total_weight = const.DEVICE_HEALTH_SCORE_CPU_WEIGHT + \
            const.DEVICE_HEALTH_SCORE_MEM_WEIGHT + \
            const.DEVICE_HEALTH_SCORE_CPS_WEIGHT

        return int((cpu_score + mem_score + cps_score) / total_weight)

    # returns percentage of TMM memory currently in use
    @log
    def get_mem_health_score(self, global_stats=None):
        """use TMM memory usage for memory health """
        if not global_stats:
            global_stats = self.get_global_statistics()
        tmm_mem = int(
            global_stats[
                'Sys::Performance System']['Memory Used'][
                'TMM Memory Used']['current']
        )
        other_mem = int(
            global_stats[
                'Sys::Performance System']['Memory Used'][
                'TMM Memory Used']['current']
        )
        if other_mem > 90:
            return other_mem
        else:
            return tmm_mem

    @log
    def get_cpu_health_score(self, global_stats=None):
        """Get cpu health score """
        if not global_stats:
            global_stats = self.get_global_statistics()
        cpu_score = int(
            global_stats[
                'Sys::Performance System']['System CPU Usage'][
                'Utilization']['current']
        )
        return cpu_score

    @log
    def get_cps_health_score(self, global_stats=None):
        """Get cps health score """
        if not global_stats:
            global_stats = self.get_global_statistics()
        count_init = int(
            global_stats[
                'Sys::Performance Connections']['Active Connections'][
                'Connections']['current']
        )
        time.sleep(const.DEVICE_HEALTH_SCORE_CPS_PERIOD)
        global_stats = self.get_global_statistics()
        count_final = int(
            global_stats[
                'Sys::Performance Connections']['Active Connections'][
                'Connections']['current']
        )
        cps = (count_final - count_init) / const.DEVICE_HEALTH_SCORE_CPS_PERIOD

        if cps >= const.DEVICE_HEALTH_SCORE_CPS_MAX:
            return 0
        else:
            score = int(100 - ((100 * float(cps)) /
                        float(const.DEVICE_HEALTH_SCORE_CPS_MAX)))
        return score

    @log
    def get_active_connection_count(self, global_stats=None):
        if not global_stats:
            global_stats = self.get_global_statistics()
        return int(
            global_stats[
                'Sys::Performance Connections'][
                'Active Connections'][
                'Connections']['current']
        )

    @log
    def get_active_SSL_TPS(self, global_stats=None):
        if not global_stats:
            global_stats = self.get_global_statistics()
        return int(
            global_stats[
                'Sys::Performance Throughput'][
                'SSL Transactions'][
                'SSL TPS']['current']
        )

    @log
    def get_inbound_throughput(self, global_stats=None):
        if not global_stats:
            global_stats = self.get_global_statistics()
        return int(
            global_stats[
                'Sys::Performance Throughput'][
                'Throughput(bits)']['In']['current']
        )

    @log
    def get_outbound_throughput(self, global_stats=None):
        if not global_stats:
            global_stats = self.get_global_statistics()
        return int(
            global_stats[
                'Sys::Performance Throughput'][
                'Throughput(bits)']['Out']['current']
        )

    @log
    def get_throughput(self, global_stats=None):
        if not global_stats:
            global_stats = self.get_global_statistics()
        inbound = int(
            global_stats[
                'Sys::Performance Throughput'][
                'Throughput(bits)']['In']['current']
        )
        outbound = int(
            global_stats[
                'Sys::Performance Throughput'][
                'Throughput(bits)']['Out']['current']
        )
        return inbound + outbound



# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class BigIPException(Exception):
    pass


class CreateAlreadyExtantResource(BigIPException):
    pass


class MinorVersionValidateFailed(BigIPException):
    pass


class MajorVersionValidateFailed(BigIPException):
    pass


class ProvisioningExtraMBValidateFailed(BigIPException):
    pass


class BigIPDeviceLockAcquireFailed(BigIPException):
    pass


class BigIPClusterInvalidHA(BigIPException):
    pass


class BigIPClusterSyncFailure(BigIPException):
    pass


class BigIPClusterPeerAddFailure(BigIPException):
    pass


class BigIPClusterConfigSaveFailure(BigIPException):
    pass


class UnknownMonitorType(BigIPException):
    pass


class MissingVTEPAddress(BigIPException):
    pass


class MissingNetwork(BigIPException):
    pass


class InvalidNetworkType(BigIPException):
    pass


class StaticARPCreationException(BigIPException):
    pass


class StaticARPQueryException(BigIPException):
    pass


class StaticARPDeleteException(BigIPException):
    pass


class ClusterCreationException(BigIPException):
    pass


class ClusterUpdateException(BigIPException):
    pass


class ClusterQueryException(BigIPException):
    pass


class ClusterDeleteException(BigIPException):
    pass


class DeviceCreationException(BigIPException):
    pass


class DeviceUpdateException(BigIPException):
    pass


class DeviceQueryException(BigIPException):
    pass


class DeviceDeleteException(BigIPException):
    pass


class InterfaceQueryException(BigIPException):
    pass


class IAppCreationException(BigIPException):
    pass


class IAppQueryException(BigIPException):
    pass


class IAppUpdateException(BigIPException):
    pass


class IAppDeleteException(BigIPException):
    pass


class L2GRETunnelCreationException(BigIPException):
    pass


class L2GRETunnelQueryException(BigIPException):
    pass


class L2GRETunnelUpdateException(BigIPException):
    pass


class L2GRETunnelDeleteException(BigIPException):
    pass


class MonitorCreationException(BigIPException):
    pass


class MonitorQueryException(BigIPException):
    pass


class MonitorUpdateException(BigIPException):
    pass


class MonitorDeleteException(BigIPException):
    pass


class NATCreationException(BigIPException):
    pass


class NATQueryException(BigIPException):
    pass


class NATUpdateException(BigIPException):
    pass


class NATDeleteException(BigIPException):
    pass


class PoolCreationException(BigIPException):
    pass


class PoolQueryException(BigIPException):
    pass


class PoolUpdateException(BigIPException):
    pass


class PoolDeleteException(BigIPException):
    pass


class RouteCreationException(BigIPException):
    pass


class RouteQueryException(BigIPException):
    pass


class RouteUpdateException(BigIPException):
    pass


class RouteDeleteException(BigIPException):
    pass


class RuleCreationException(BigIPException):
    pass


class RuleQueryException(BigIPException):
    pass


class RuleUpdateException(BigIPException):
    pass


class RuleDeleteException(BigIPException):
    pass


class SelfIPCreationException(BigIPException):
    pass


class SelfIPQueryException(BigIPException):
    pass


class SelfIPUpdateException(BigIPException):
    pass


class SelfIPDeleteException(BigIPException):
    pass


class SNATCreationException(BigIPException):
    pass


class SNATQueryException(BigIPException):
    pass


class SNATUpdateException(BigIPException):
    pass


class SNATDeleteException(BigIPException):
    pass


class SystemCreationException(BigIPException):
    pass


class SystemQueryException(BigIPException):
    pass


class SystemUpdateException(BigIPException):
    pass


class SystemDeleteException(BigIPException):
    pass


class VirtualServerCreationException(BigIPException):
    pass


class VirtualServerQueryException(BigIPException):
    pass


class VirtualServerUpdateException(BigIPException):
    pass


class VirtualServerDeleteException(BigIPException):
    pass


class VLANCreationException(BigIPException):
    pass


class VLANQueryException(BigIPException):
    pass


class VLANUpdateException(BigIPException):
    pass


class VLANDeleteException(BigIPException):
    pass


class VXLANCreationException(BigIPException):
    pass


class VXLANQueryException(BigIPException):
    pass


class VXLANUpdateException(BigIPException):
    pass


class VXLANDeleteException(BigIPException):
    pass



# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import sys


class Log(object):
    @staticmethod
    def debug(prefix, msg):
        Log._log('debug', prefix, msg)

    @staticmethod
    def error(prefix, msg):
        Log._log('error', prefix, msg)

    @staticmethod
    def crit(prefix, msg):
        Log._log('crit', prefix, msg)

    @staticmethod
    def info(prefix, msg):
        Log._log('info', prefix, msg)

    @staticmethod
    def _log(level, prefix, msg):
        log_string = prefix + ': ' + msg
        log = logging.getLogger(__name__)
        out_hdlr = logging.StreamHandler(sys.stdout)
        out_hdlr.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
        log.addHandler(out_hdlr)

        if level == 'debug':
            log.debug(log_string)
        elif level == 'error':
            log.error(log_string)
        elif level == 'crit':
            log.critical(log_string)
        else:
            log.info(log_string)

        log.removeHandler(out_hdlr)



"""vxlan.py """
# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=broad-except

from eventlet import greenthread
from f5.bigip import exceptions
from f5.bigip.rest_collection import icontrol_rest_folder
from f5.bigip.rest_collection import log
from f5.bigip.rest_collection import prefixed
from f5.bigip.rest_collection import strip_folder_and_prefix
from f5.common import constants as const
from f5.common.logger import Log

import json
import os


class VXLAN(object):
    """Interface for vxlan related REST methods """

    OBJ_PREFIX = 'uuid_'

    def __init__(self, bigip):
        self.bigip = bigip

    @icontrol_rest_folder
    @log
    def create_multipoint_profile(self, name=None, folder='Common'):
        """Create vxlan multipoint tunnel profile """
        if not self.profile_exists(name=name, folder=folder):
            folder = str(folder).replace('/', '')
            payload = dict()
            payload['name'] = name
            payload['partition'] = folder
            payload['defaultsFrom'] = 'vxlan'
            payload['floodingType'] = 'multipoint'
            payload['port'] = const.VXLAN_UDP_PORT
            request_url = self.bigip.icr_url + '/net/tunnels/vxlan/'
            response = self.bigip.icr_session.post(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            elif response.staus_code == 409:
                return True
            else:
                Log.error('VXLAN', response.text)
                raise exceptions.VXLANCreationException(response.text)
        else:
            return False

    @icontrol_rest_folder
    @log
    def delete_profile(self, name=None, folder='Common'):
        """Delete vxlan multipoint tunnel profile """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/tunnels/vxlan/'
        request_url += '~' + folder + '~' + name

        response = self.bigip.icr_session.delete(
            request_url, timeout=const.CONNECTION_TIMEOUT)

        if response.status_code < 400:
            return True
        elif response.status_code == 404:
            return True
        else:
            Log.error('VXLAN', response.text)
            raise exceptions.VXLANDeleteException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def create_multipoint_tunnel(self, name=None,
                                 profile_name=None,
                                 self_ip_address=None,
                                 vxlanid=0,
                                 description=None,
                                 folder='Common',
                                 route_domain_id=0):
        """Create vxlan multipoint tunnel """
        if not self.tunnel_exists(name=name, folder=folder):
            folder = str(folder).replace('/', '')
            # check partition is okay to create in
            payload = dict()
            payload['name'] = name
            payload['partition'] = folder
            payload['profile'] = profile_name
            payload['key'] = vxlanid
            payload['localAddress'] = self_ip_address
            payload['remoteAddress'] = '0.0.0.0'
            if description:
                payload['description'] = description
            request_url = self.bigip.icr_url + '/net/tunnels/tunnel/'
            response = self.bigip.icr_session.post(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                if not folder == 'Common':
                    self.bigip.route.add_vlan_to_domain_by_id(
                        name=name, folder=folder,
                        route_domain_id=route_domain_id)
                return True
            else:
                Log.error('VXLAN', response.text)
                raise exceptions.VXLANCreationException(response.text)
        else:
            return False

    @icontrol_rest_folder
    @log
    def delete_tunnel(self, name=None, folder='Common'):
        """Delete vxlan multipoint tunnel """
        folder = str(folder).replace('/', '')
        # delete arp and fdb records for this tunnel first
        request_url = self.bigip.icr_url + '/net/fdb/tunnel/'
        request_url += '~' + folder + '~' + name + '?ver=11.5.0'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if const.FDB_POPULATE_STATIC_ARP:
                if 'records' in response_obj:
                    for record in response_obj['records']:
                        self.bigip.arp.delete_by_mac(
                            mac_address=record['name'], folder=folder)
            payload = dict()
            payload['records'] = []
            tunnel_link = self.bigip.icr_link(response_obj['selfLink'])
            response = self.bigip.icr_session.patch(
                tunnel_link, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code > 399:
                Log.error('fdb', response.text)
                raise exceptions.VXLANDeleteException(response.text)
        elif response.status_code != 404:
            Log.error('fdb', response.text)
            raise exceptions.VXLANQueryException(response.text)
        request_url = self.bigip.icr_url + '/net/tunnels/tunnel/'
        request_url += '~' + folder + '~' + name
        response = self.bigip.icr_session.delete(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        elif response.status_code == 404:
            return True
        else:
            Log.error('VXLAN', response.text)
            raise exceptions.VXLANDeleteException(response.text)

    @icontrol_rest_folder
    @log
    def delete_all(self, folder='Common'):
        """Delete vxlan multipoint tunnels """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/tunnels/tunnel/'
        request_url += '?$select=name,selfLink'
        request_filter = 'partition eq ' + folder
        request_url += '&$filter=' + request_filter
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for item in response_obj['items']:
                    if item['name'].startswith(self.OBJ_PREFIX):
                        self.delete_all_fdb_entries(item['name'], folder)
                        response = self.bigip.icr_session.delete(
                            self.bigip.icr_link(item['selfLink']),
                            timeout=const.CONNECTION_TIMEOUT)
                        if response.status_code > 400 and \
                           response.status_code != 404:
                            Log.error('VXLAN', response.text)
                            raise exceptions.VXLANDeleteException(
                                response.text)
        elif response.status_code != 404:
            Log.error('VXLAN', response.text)
            raise exceptions.VXLANQueryException(response.text)
        return True

    @icontrol_rest_folder
    @log
    def get_fdb_entry(self,
                      tunnel_name=None,
                      mac=None,
                      folder='Common'):
        """Get vxlan fdb entry """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/fdb/tunnel/'
        request_url += '~' + folder + '~' + tunnel_name + '?ver=11.5.0'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'records' in response_obj:
                if not mac:
                    return response_obj['records']
                else:
                    for record in response_obj['records']:
                        if record['name'] == mac:
                            return record
        elif response.status_code != 404:
            Log.error('VXLAN', response.text)
            raise exceptions.VXLANQueryException(response.text)
        return []

    @icontrol_rest_folder
    @log
    def add_fdb_entry(self,
                      tunnel_name=None,
                      mac_address=None,
                      vtep_ip_address=None,
                      arp_ip_address=None,
                      folder=None):
        """Add vxlan fdb entry """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/fdb/tunnel/'
        request_url += '~' + folder + '~' + tunnel_name + '?ver=11.5.0'
        records = self.get_fdb_entry(tunnel_name=tunnel_name,
                                     mac=None,
                                     folder=folder)
        fdb_entry = dict()
        fdb_entry['name'] = mac_address
        fdb_entry['endpoint'] = vtep_ip_address

        for i in range(len(records)):
            if records[i]['name'] == mac_address:
                records[i] = fdb_entry
                break
        else:
            records.append(fdb_entry)

        payload = dict()
        payload['records'] = records
        response = self.bigip.icr_session.patch(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            if const.FDB_POPULATE_STATIC_ARP:
                if arp_ip_address:
                    try:
                        if self.bigip.arp.create(ip_address=arp_ip_address,
                                                 mac_address=mac_address,
                                                 folder=folder):
                            return True
                        else:
                            return False
                    except Exception as exc:
                        Log.error('VXLAN',
                                  'could not create static arp: %s on %s'
                                  % (exc.message, self.bigip.device_name))
                        return False
            return True
        else:
            Log.error('VXLAN', response.text)
            raise exceptions.VXLANUpdateException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def add_fdb_entries(self, fdb_entries=None):
        """Add vxlan fdb entries """
        for tunnel_name in fdb_entries:
            folder = fdb_entries[tunnel_name]['folder']
            if folder != 'Common':
                folder = prefixed(folder)
            request_url = self.bigip.icr_url + '/net/fdb/tunnel/'
            request_url += '~' + folder + '~' + self.OBJ_PREFIX + \
                tunnel_name + '?ver=11.5.0'
            existing_records = self.get_fdb_entry(tunnel_name=tunnel_name,
                                                  mac=None,
                                                  folder=folder)
            new_records = []
            new_mac_addresses = []
            new_arp_addresses = {}

            tunnel_records = fdb_entries[tunnel_name]['records']
            for mac in tunnel_records:
                fdb_entry = dict()
                fdb_entry['name'] = mac
                fdb_entry['endpoint'] = tunnel_records[mac]['endpoint']
                new_records.append(fdb_entry)
                new_mac_addresses.append(mac)
                if tunnel_records[mac]['ip_address']:
                    new_arp_addresses[mac] = tunnel_records[mac]['ip_address']

            for record in existing_records:
                if not record['name'] in new_mac_addresses:
                    new_records.append(record)
                else:
                    # This fdb entry exists and is not being updated.
                    # So, do not update the ARP record either.
                    if record['name'] in new_arp_addresses:
                        del new_arp_addresses[record['name']]

            payload = dict()
            payload['records'] = new_records
            response = self.bigip.icr_session.patch(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                if const.FDB_POPULATE_STATIC_ARP:
                    for mac in new_arp_addresses:
                        try:
                            self.bigip.arp.create(
                                ip_address=new_arp_addresses[mac],
                                mac_address=mac,
                                folder=folder)
                        except Exception as exc:
                            Log.error('VXLAN',
                                      'could not create static arp: %s'
                                      % exc.message)
            return True
        return False

    @icontrol_rest_folder
    @log
    def delete_fdb_entry(self,
                         tunnel_name=None,
                         mac_address=None,
                         arp_ip_address=None,
                         folder='Common'):
        """Delete vxlan fdb entry """
        folder = str(folder).replace('/', '')
        if const.FDB_POPULATE_STATIC_ARP:
            if arp_ip_address:
                self.bigip.arp.delete(ip_address=arp_ip_address,
                                      folder=folder)
        request_url = self.bigip.icr_url + '/net/fdb/tunnel/'
        request_url += '~' + folder + '~' + tunnel_name + '?ver=11.5.0'
        records = self.get_fdb_entry(tunnel_name=tunnel_name,
                                     mac=None,
                                     folder=folder)
        if not records:
            return False
        original_len = len(records)
        records = [record for record in records
                   if record.get('name') != mac_address]
        if original_len != len(records):
            if len(records) == 0:
                records = None
            payload = dict()
            payload['records'] = records
            response = self.bigip.icr_session.patch(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            elif response.status_code == 404:
                return True
            else:
                Log.error('VXLAN', response.text)
                raise exceptions.VXLANUpdateException(response.text)
            return False
        return False

    @icontrol_rest_folder
    @log
    def delete_fdb_entries(self,
                           tunnel_name=None,
                           fdb_entries=None):
        """Delete vxlan fdb entries """
        for tunnel_name in fdb_entries:
            folder = fdb_entries[tunnel_name]['folder']
            if folder != 'Common':
                folder = prefixed(folder)
            request_url = self.bigip.icr_url + '/net/fdb/tunnel/'
            request_url += '~' + folder + '~' + tunnel_name + '?ver=11.5.0'
            existing_records = self.get_fdb_entry(tunnel_name=tunnel_name,
                                                  mac=None,
                                                  folder=folder)
            arps_to_delete = {}
            new_records = []

            for record in existing_records:
                for mac in fdb_entries[tunnel_name]['records']:
                    if record['name'] == mac and mac['ip_address']:
                        arps_to_delete[mac] = mac['ip_address']
                        break
                else:
                    new_records.append(record)

            if len(new_records) == 0:
                new_records = None
            payload = dict()
            payload['records'] = new_records
            response = self.bigip.icr_session.patch(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                if const.FDB_POPULATE_STATIC_ARP:
                    for mac in arps_to_delete:
                        self.bigip.arp.delete(ip_address=arps_to_delete[mac],
                                              folder='Common')
            return True
        return False

    @icontrol_rest_folder
    @log
    def delete_all_fdb_entries(self, tunnel_name=None, folder='Common'):
        """Delete fdb entries """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/fdb/tunnel/'
        request_url += '~' + folder + '~' + tunnel_name + '?ver=11.5.0'
        response = self.bigip.icr_session.patch(
            request_url, data=json.dumps({'records': None}),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400 or response.status_code == 404:
            return True
        else:
            Log.error('VXLAN', response.text)
            raise exceptions.VXLANUpdateException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def get_profiles(self, folder='Common'):
        """Get tunnel profiles """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/tunnels/vxlan'
        request_filter = 'partition eq ' + folder
        request_url += '?$filter=' + request_filter
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return_obj = json.loads(response.text)
            profile_list = []
            if 'items' in return_obj:
                for profile in return_obj['items']:
                    profile_list.append(
                        strip_folder_and_prefix(profile['name']))
                return profile_list
            else:
                return None
        elif response.status_code != 404:
            Log.error('VXLAN', response.text)
            raise exceptions.VXLANQueryException(response.text)
        return None

    @icontrol_rest_folder
    @log
    def profile_exists(self, name=None, folder='Common'):
        """Does vxlan tunnel exist? """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/tunnels/vxlan/'
        request_url += '~' + folder + '~' + name

        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        elif response.status_code != 404:
            Log.error('VXLAN', response.text)
            raise exceptions.VXLANQueryException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def get_tunnels(self, folder='Common'):
        """Get tunnels """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/tunnels/tunnel'
        if folder:
            request_filter = 'partition eq ' + folder
            request_url += '?$filter=' + request_filter
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            tunnel_list = []
            return_obj = json.loads(response.text)
            if 'items' in return_obj:
                for tunnel in return_obj['items']:
                    if tunnel['profile'].find('vxlan') > 0:
                        tunnel_list.append(
                            strip_folder_and_prefix(tunnel['name']))
                return tunnel_list
            else:
                return None
        elif response.status_code != 404:
            Log.error('VXLAN', response.text)
            exceptions.VXLANQueryException(response.text)
        return None

    @icontrol_rest_folder
    @log
    def get_tunnel_key(self, name=None, folder='Common'):
        """Get tunnel key """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/tunnels/tunnel/'
        request_url += '~' + folder + '~' + name
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return_obj = json.loads(response.text)
            Log.debug('VXLAN', 'get_tunnel_key got %s' + str(return_obj))
            return return_obj['key']
        elif response.status_code != 404:
            Log.error('VXLAN', response.text)
            exceptions.VXLANQueryException(response.text)
        return None

    @icontrol_rest_folder
    @log
    def get_tunnel_by_description(self, description=None, folder='Common'):
        """Get tunnel by description """
        folder = str(folder).replace('/', '')
        if description:
            request_url = self.bigip.icr_url + '/net/tunnels/tunnel/'
            if folder:
                request_filter = 'partition eq ' + folder
                request_url += '?$filter=' + request_filter
            response = self.bigip.icr_session.get(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return_obj = json.loads(response.text)
                if 'items' in return_obj:
                    for tunnel in return_obj['items']:
                        if tunnel['description'] == description:
                            return strip_folder_and_prefix(tunnel['name'])
                return None
            elif response.status_code != 404:
                Log.error('VXLAN', response.text)
                raise exceptions.VXLANQueryException(response.text)
        return None

    @icontrol_rest_folder
    @log
    def get_tunnel_folder(self, tunnel_name=None):
        """Get tunnel folder """
        if tunnel_name:
            request_url = self.bigip.icr_url + '/net/tunnels/tunnel/'
            response = self.bigip.icr_session.get(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return_obj = json.loads(response.text)
                if 'items' in return_obj:
                    for tunnel in return_obj['items']:
                        if tunnel['name'] == tunnel_name:
                            return tunnel['partition']
                return None
            elif response.status_code != 404:
                Log.error('VXLAN', response.text)
                raise exceptions.VXLANQueryException(response.text)
        return None

    @icontrol_rest_folder
    @log
    def tunnel_exists(self, name=None, folder='Common'):
        """Does tunnel exist? """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/tunnels/tunnel/'
        request_url += '~' + folder + '~' + name

        for retry in range(2):
            if retry > 0:
                Log.error('VXLAN', 'Attempting REST retry after 401 error')
            response = self.bigip.icr_session.get(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            elif response.status_code == 401:
                if retry < 1:
                    greenthread.sleep(1)
                    continue
                else:
                    Log.error('VXLAN', response.text)
                    raise exceptions.VXLANQueryException(response.text)
            elif response.status_code != 404:
                Log.error('VXLAN', response.text)
                raise exceptions.VXLANQueryException(response.text)
            else:
                return False
        return False

    @icontrol_rest_folder
    def _in_use(self, name=None, folder=None):
        """Is tunnel used by selfip? """
        if name:
            folder = str(folder).replace('/', '')
            request_url = self.bigip.icr_url + '/net/self?$select=vlan'
            if folder:
                request_filter = 'partition eq ' + folder
                request_url += '&$filter=' + request_filter
            else:
                folder = 'Common'
            response = self.bigip.icr_session.get(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return_obj = json.loads(response.text)
                if 'items' in return_obj:
                    for selfip in return_obj['items']:
                        vlan_name = os.path.basename(selfip['vlan'])
                        if vlan_name == name:
                            return True
                        if vlan_name == \
                           strip_folder_and_prefix(name):
                            return True
            elif response.status_code != 404:
                Log.error('self', response.text)
                raise exceptions.VXLANQueryException(response.text)
        return False



# Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class MemberStateAlwaysRequiredOnUpdate(Exception):
    pass


class PoolCollection(Collection):
    def __init__(self, ltm):
        super(PoolCollection, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Pool]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:pool:poolstate': Pool}


class Pool(Resource):
    def __init__(self, pool_collection):
        super(Pool, self).__init__(pool_collection)
        self._meta_data['required_json_kind'] = 'tm:ltm:pool:poolstate'
        self._meta_data['attribute_registry'] = {
            'tm:ltm:pool:memberscollectionstate': MembersCollection
        }


class MembersCollection(Collection):
    def __init__(self, pool):
        super(MembersCollection, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Member]
        self._meta_data['required_json_kind'] =\
            'tm:ltm:pool:members:memberscollectionstate'
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:pool:members:membersstate': Member}


class Member(Resource):
    def __init__(self, member_collection):
        super(Member, self).__init__(member_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:pool:members:membersstate'
        self._meta_data['required_creation_parameters'].update(('partition',))

    def update(self, **kwargs):
        try:
            state = kwargs.pop('state')
        except KeyError:
            error_message = 'You must supply a value to the "state"' +\
                ' parameter if you do not wish to change the state then' +\
                ' pass "state=None".'
            raise MemberStateAlwaysRequiredOnUpdate(error_message)
        if state is None:
            self.__dict__.pop(u'state', '')
        else:
            self.state = state
        # This is an example implementation of read-only params
        self.__dict__.pop(u'ephemeral', '')
        self.__dict__.pop(u'address', '')
        self._update(**kwargs)



# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# NOTE:  Code taken from Effective Python Item 26


class ToDictMixin(object):
    traversed = {}
    Containers = tuple, list, set, frozenset, dict

    def to_dict(self):
        ToDictMixin.traversed = {}
        return self._to_dict()

    def _to_dict(self):
        result = self._traverse_dict(self.__dict__)
        return result

    def _traverse_dict(self, instance_dict):
        output = {}
        for key, value in instance_dict.items():
            output[key] = self._traverse(key, value)
        return output

    def _traverse(self, key, value):
        if isinstance(value, ToDictMixin.Containers) or\
           hasattr(value, '__dict__'):
            if id(value) in ToDictMixin.traversed:
                return ToDictMixin.traversed[id(value)]
            else:
                ToDictMixin.traversed[id(value)] = ['TraversalRecord', key]
        if isinstance(value, ToDictMixin):
            return value._to_dict()
        elif isinstance(value, dict):
            return self._traverse_dict(value)
        elif isinstance(value, list):
            return [self._traverse(key, item) for item in value]
        elif hasattr(value, '__dict__'):
            return self._traverse_dict(value.__dict__)
        else:
            return value


class LazyAttributesRequired(Exception):
    pass


class LazyAttributeMixin(object):
    def __getattr__(self, name):
        # ensure this object supports lazy attrs.
        cls_name = self.__class__.__name__
        if '_meta_data' not in self.__dict__:
            error_message = '%r does not have self._meta_data' % cls_name
            raise LazyAttributesRequired(error_message)
        elif 'allowed_lazy_attributes' not in self._meta_data:
            error_message = ('"allowed_lazy_attributes" not in',
                             'self._meta_data for class %s' % cls_name)
            raise LazyAttributesRequired(error_message)

        # ensure the requested attr is present
        lower_attr_names =\
            [la.__name__.lower() for la in
                self._meta_data['allowed_lazy_attributes']]
        if name not in lower_attr_names:
            error_message = "'%s' object has no attribute '%s'"\
                % (self.__class__, name)
            raise AttributeError(error_message)

        # Instantiate and potentially set the attr on the object
        # Issue #112 -- Only call setattr here if the lazy attribute
        # is NOT a `Resource`.  This should allow for only 1 ltm attribute
        # but many nat attributes just like the BIGIP device.
        for lazy_attribute in self._meta_data['allowed_lazy_attributes']:
            if name == lazy_attribute.__name__.lower():
                iface_collection = lazy_attribute(self)
                # Use the name of ResourceResource because importing causes
                # a circular reference
                bases = [base.__name__ for base in lazy_attribute.__bases__]
                if 'Resource' not in bases:
                    setattr(self, name, iface_collection)
                return iface_collection


class ExclusiveAttributesMixin(object):
    def __setattr__(self, key, value):
        '''Remove any of the existing exclusive attrs from the object

        Objects attributes can be exclusive for example disable/enable.  So
        we need to make sure objects only have one of these attributes at
        at time so that the updates won't fail.
        '''
        if '_meta_data' in self.__dict__:
            # Sometimes this is called prior to full object construction
            for attr_set in self._meta_data['exclusive_attributes']:
                if key in attr_set:
                    new_set = set(attr_set) - set([key])
                    [self.__dict__.pop(n, '') for n in new_set]
        # Now set the attribute
        super(ExclusiveAttributesMixin, self).__setattr__(key, value)



# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip import exceptions
from f5.bigip.rest_collection import log
from f5.common import constants as const
from f5.common.logger import Log

import base64
import json
import os
import time


# Management - Cluster
class Cluster(object):
    def __init__(self, bigip):
        self.bigip = bigip

        self.bigip.icontrol.add_interfaces(['Management.Trust'])
        self.mgmt_trust = self.bigip.icontrol.Management.Trust

    @log
    def get_sync_status(self):
        """Get the sync status description for the bigip """
        request_url = self.bigip.icr_url + '/cm/sync-status?$select=status'
        response = self.bigip.icr_session.get(request_url,
                                              timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            entries = response_obj['entries']
            status = entries['https://localhost/mgmt/tm/cm/sync-status/0']
            desc = status['nestedStats']['entries']['status']['description']
            return desc
        else:
            Log.error('sync-status', response.text)
            raise exceptions.ClusterQueryException(response.text)
        return None

    @log
    def get_sync_color(self):
        """Get the sync color for the bigip """
        request_url = self.bigip.icr_url + '/cm/sync-status?$select=color'
        response = self.bigip.icr_session.get(request_url,
                                              timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            entries = response_obj['entries']
            status = entries['https://localhost/mgmt/tm/cm/sync-status/0']
            desc = status['nestedStats']['entries']['color']['description']
            return desc
        else:
            Log.error('sync-info', response.text)
            raise exceptions.ClusterQueryException(response.text)
        return None

    @log
    def save_config(self):
        """Save the bigip configuration """
        request_url = self.bigip.icr_url + '/sys/config'
        payload = dict()
        payload['command'] = 'save'
        response = self.bigip.icr_session.post(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            Log.error('config', response.text)
            raise exceptions.BigIPClusterConfigSaveFailure(response.text)
        return False

    @log
    def get_local_device_name(self):
        """Get local device name """
        request_url = self.bigip.icr_url + '/cm/device'
        request_url += '?$select=selfDevice,name,hostname,managementIp'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for device in response_obj['items']:
                    if device['selfDevice'].lower() == 'true':
                        return device['name']
        else:
            Log.error('device', response.text)
            raise exceptions.ClusterQueryException(response.text)
        return None

    @log
    def get_local_device_addr(self):
        """Get local device management ip """
        request_url = self.bigip.icr_url + '/cm/device'
        request_url += '?$select=selfDevice,name,hostname,managementIp'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for device in response_obj['items']:
                    if device['selfDevice'].lower() == 'true':
                        return device['managementIp']
        else:
            Log.error('device', response.text)
            raise exceptions.ClusterQueryException(response.text)
        return None

    @log
    def sync_local_device_to_group(self, device_group_name):
        """Sync local device to group """
        request_url = self.bigip.icr_url + '/cm'
        payload = dict()
        payload['command'] = 'run'
        payload['options'] = [{'config-sync': 'to-group ' + device_group_name}]
        response = self.bigip.icr_session.post(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            Log.error('sync', response.text)
            raise exceptions.BigIPClusterSyncFailure(response.text)
        return False

    # force_now=True is typically used for initial sync.
    # In order to avoid sync problems, you should wait until devices
    # in the group are connected.
    @log
    def sync(self, name, force_now=False):
        """Ensure local device in sync with group """
        sync_start_time = time.time()
        dev_name = self.get_local_device_name()
        sleep_delay = const.SYNC_DELAY

        attempts = 0
        if force_now:
            self.sync_local_device_to_group(name)
            time.sleep(sleep_delay)
            attempts += 1

        while attempts < const.MAX_SYNC_ATTEMPTS:
            state = self.get_sync_status()
            if state in ['Standalone', 'In Sync']:
                break

            elif state == 'Awaiting Initial Sync':
                attempts += 1
                Log.info(
                    'Cluster',
                    "Device %s - Synchronizing initial config to group %s"
                    % (dev_name, name))
                self.sync_local_device_to_group(name)
                time.sleep(sleep_delay)

            elif state in ['Disconnected',
                           'Not All Devices Synced',
                           'Changes Pending']:
                attempts += 1

                last_log_time = 0
                now = time.time()
                wait_start_time = now
                # Keep checking the sync state in a quick loop.
                # We want to detect In Sync as quickly as possible.
                while now - wait_start_time < sleep_delay:
                    # Only log once per second
                    if now - last_log_time >= 1:
                        Log.info(
                            'Cluster',
                            'Device %s, Group %s not synced. '
                            % (dev_name, name) +
                            'Waiting. State is: %s'
                            % state)
                        last_log_time = now
                    state = self.get_sync_status()
                    if state in ['Standalone', 'In Sync']:
                        break
                    time.sleep(.5)
                    now = time.time()
                else:
                    # if we didn't break out due to the group being in sync
                    # then attempt to force a sync.
                    self.sync_local_device_to_group(name)
                    sleep_delay += const.SYNC_DELAY
                    # no need to sleep here because we already spent the sleep
                    # interval checking status.
                    continue

                # Only a break from the inner while loop due to Standalone or
                # In Sync will reach here.
                # Normal exit of the while loop reach the else statement
                # above which continues the outer loop
                break

            elif state == 'Sync Failure':
                Log.info('Cluster',
                         "Device %s - Synchronization failed for %s"
                         % (dev_name, name))
                Log.debug('Cluster', 'SYNC SECONDS (Sync Failure): ' +
                          str(time.time() - sync_start_time))
                raise exceptions.BigIPClusterSyncFailure(
                    'Device service group %s' % name +
                    ' failed after ' +
                    '%s attempts.' % const.MAX_SYNC_ATTEMPTS +
                    ' Correct sync problem manually' +
                    ' according to sol13946 on ' +
                    ' support.f5.com.')
            else:
                attempts += 1
                Log.info('Cluster',
                         "Device %s " % dev_name +
                         "Synchronizing config attempt %s to group %s:"
                         % (attempts, name) + " current state: %s" % state)
                self.sync_local_device_to_group(name)
                time.sleep(sleep_delay)
                sleep_delay += const.SYNC_DELAY
        else:
            if state == 'Disconnected':
                Log.debug('Cluster',
                          'SYNC SECONDS(Disconnected): ' +
                          str(time.time() - sync_start_time))
                raise exceptions.BigIPClusterSyncFailure(
                    'Device service group %s' % name +
                    ' could not reach a sync state' +
                    ' because they can not communicate' +
                    ' over the sync network. Please' +
                    ' check connectivity.')
            else:
                Log.debug('Cluster', 'SYNC SECONDS(Timeout): ' +
                          str(time.time() - sync_start_time))
                raise exceptions.BigIPClusterSyncFailure(
                    'Device service group %s' % name +
                    ' could not reach a sync state after ' +
                    '%s attempts.' % const.MAX_SYNC_ATTEMPTS +
                    ' It is in %s state currently.' % state +
                    ' Correct sync problem manually' +
                    ' according to sol13946 on ' +
                    ' support.f5.com.')

        Log.debug('Cluster', 'SYNC SECONDS(Success): ' +
                  str(time.time() - sync_start_time))

    @log
    def sync_failover_dev_group_exists(self, name):
        """Does the sync failover device group exist? """
        request_url = self.bigip.icr_url + '/cm/device-group/'
        request_url += '~Common~' + name
        request_url += '?$select=type'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'type' in response_obj:
                if response_obj['type'] == 'sync-failover':
                    return True
                else:
                    return False
        elif response.status_code == 404:
            return False
        else:
            Log.error('device-group', response.text)
            raise exceptions.ClusterQueryException(response.text)
        return False

    @log
    def wait_for_insync_status(self):
        """Wait until sync status is 'in sync' and color is 'green'. """
        sync_status_attempts = 0
        max_status_attempts = 60
        while sync_status_attempts < max_status_attempts:
            sync_color = self.get_sync_color().lower()
            sync_status = self.get_sync_status().lower()
            if sync_color != u'green' and sync_status != u'in sync':
                sync_status_attempts += 1
                time.sleep(2)
                continue
            return
        raise exceptions.BigIPClusterPeerAddFailure(
            'Group failed to sync in 60 seconds while adding peer.'
        )

    @log
    def add_peer(self, name, mgmt_ip_address, username, password):
        """Add a peer to the local trust group """
        if not self.peer_exists(name):
            if self.bigip.device.get_lock():
                local_device = self.get_local_device_name()
                local_mgmt_address = self.get_local_device_addr()
                root_mgmt_dict = {'root_device_name': local_device,
                                  'root_device_mgmt_address':
                                  local_mgmt_address}
                local_md = self.bigip.device.get_metadata()
                if local_md and 'root_device_name' in local_md.keys():
                    md_device_name = os.path.basename(
                        local_md['root_device_name'])
                    if md_device_name:
                        if not md_device_name == local_device:
                            raise exceptions.BigIPClusterPeerAddFailure(
                                'the device used to peer %s ' % name +
                                ' was already itself peered from root' +
                                ' device: %s'
                                % local_md['root_device_name'])
                self.bigip.device.update_metadata(None, root_mgmt_dict)
                Log.info('Cluster', 'Device %s - adding peer %s'
                                    % (local_device, name))

                self.mgmt_trust.add_authority_device(mgmt_ip_address,
                                                     username,
                                                     password,
                                                     name,
                                                     '', '',
                                                     '', '')
                attempts = 0
                while attempts < const.PEER_ADD_ATTEMPTS_MAX:
                    if self.get_sync_status() == "OFFLINE":
                        self.mgmt_trust.remove_device([name])
                        self.mgmt_trust.add_authority_device(mgmt_ip_address,
                                                             username,
                                                             password,
                                                             name,
                                                             '', '',
                                                             '', '')
                    else:
                        self.wait_for_insync_status()
                        self.bigip.device.release_lock()
                        return
                    time.sleep(const.PEER_ADD_ATTEMPT_DELAY)
                    attempts += 1
                else:
                    raise exceptions.BigIPClusterPeerAddFailure(
                        'Could not add peer device %s' % name +
                        ' as a trust for device %s'
                        % os.path.basename(self.get_local_device_name()) +
                        ' after % attempts' % const.PEER_ADD_ATTEMPTS_MAX)
            else:
                raise exceptions.BigIPDeviceLockAcquireFailed(
                    'Unable to obtain device lock for device %s'
                    % os.path.basename(self.get_local_device_name())
                    )

    @log
    def get_peer_addr(self, name):
        """Get a peer management ip """
        request_url = self.bigip.icr_url + '/cm/device'
        request_url += '?$select=selfDevice,name,hostname,managementIp'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for device in response_obj['items']:
                    if device['name'] == name:
                        return device['managementIp']
        else:
            Log.error('device', response.text)
            raise exceptions.ClusterQueryException(response.text)
        return None

    @log
    def peer_exists(self, name):
        """Does a peer exist by name? """
        request_url = self.bigip.icr_url + '/cm/device'
        request_url += '?$select=selfDevice,name,hostname,managementIp'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for device in response_obj['items']:
                    if device['name'] == name:
                        return True
        elif response.status_code == 404:
            return False
        else:
            Log.error('device', response.text)
        return False

    @log
    def cluster_exists(self, name):
        """Does a cluster exist by name? """
        request_url = self.bigip.icr_url + '/cm/device-group/~Common~'
        request_url += name
        request_filter = '/?$select=name,type'
        request_url += request_filter
        response = self.bigip.icr_session.get(
            request_url, data=None, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if response_obj['type'] == 'sync-failover':
                return True
            else:
                return False
        elif response.status_code == 404:
            return False
        else:
            Log.error('device-group', response.text)
            return False

    @log
    def create(self, name, autosync=True):
        """Create a device group """
        request_url = self.bigip.icr_url + '/cm/device-group'
        payload = dict()
        payload['name'] = name
        if autosync:
            payload['autoSync'] = 'enabled'
        payload['networkFailover'] = 'enabled'
        payload['type'] = 'sync-failover'
        response = self.bigip.icr_session.post(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        elif response.status_code == 409:
            return True
        else:
            Log.error('device-group', response.text)
            raise exceptions.ClusterCreationException(response.text)

    @log
    def delete(self, name):
        """Delete a device group """
        if self.cluster_exists(name):
            self.remove_all_devices(name)
        request_url = self.bigip.icr_url + '/cm/device-group/~Common~'
        request_url += name
        response = self.bigip.icr_session.delete(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        elif response.status_code == 404:
            return True
        else:
            Log.error('device-group', response.text)
            raise exceptions.ClusterDeleteException(response.text)

    @log
    def enable_auto_sync(self, name):
        """Enable autosync on a device group """
        payload = dict()
        payload['autoSync'] = 'enabled'
        request_url = self.bigip.icr_url + '/cm/device-group/~Common~'
        request_url += name
        response = self.bigip.icr_session.put(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            Log.error('device-group', response.text)
            raise exceptions.ClusterUpdateException(response.text)

    @log
    def disable_auto_sync(self, name):
        """Disable autosync on a device group """
        payload = dict()
        payload['autoSync'] = 'disabled'
        request_url = self.bigip.icr_url + '/cm/device-group/~Common~'
        request_url += name
        response = self.bigip.icr_session.put(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            Log.error('device-group', response.text)
            raise exceptions.ClusterUpdateException(response.text)

    @log
    def devices(self, name):
        """Get device group devices """
        request_url = self.bigip.icr_url + '/cm/device-group/~Common~'
        request_url += name
        request_url += "/devices?$select=name"
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        return_devices = []
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for device in response_obj['items']:
                    return_devices.append(device['name'])
            return return_devices
        if response.status_code == 404:
            return []
        else:
            Log.error('device-group', response.text)
            raise exceptions.ClusterQueryException(response.text)

    @log
    def add_devices(self, name, device_names):
        """Add devices to device group """
        existing_devices = self.devices(name)
        if not isinstance(device_names, list):
            device_names = [device_names]
        need_to_update = False
        for device in device_names:
            if device not in existing_devices:
                existing_devices.append(device)
                need_to_update = True
        if need_to_update:
            request_url = self.bigip.icr_url + '/cm/device-group/~Common~'
            request_url += name
            payload = dict()
            devices_list = list()
            for device in existing_devices:
                devices_list.append({'name': device})
            payload['devices'] = devices_list
            response = self.bigip.icr_session.put(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            elif response.status_code == 409:
                return True
            else:
                Log.error('device-group', response.text)
                exceptions.ClusterUpdateException(response.text)

    @log
    def remove_devices(self, name, device_names):
        """Remove devices from device group """
        existing_devices = self.devices(name)
        if not isinstance(device_names, list):
            device_names = [device_names]
        need_to_update = False
        for device in device_names:
            if device in existing_devices:
                existing_devices.remove(device)
                need_to_update = True
        if need_to_update:
            request_url = self.bigip.icr_url + '/cm/device-group/~Common~'
            request_url += name
            payload = dict()
            devices_list = list()
            for device in existing_devices:
                devices_list.append({'name': device})
            payload['devices'] = devices_list
            response = self.bigip.icr_session.put(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            elif response.status_code == 404:
                return True
            else:
                Log.error('device-group', response.text)
                raise exceptions.ClusterUpdateException(response.text)

    @log
    def remove_all_devices(self, name):
        """Remove all devices from device group """
        request_url = self.bigip.icr_url + '/cm/device-group/~Common~'
        request_url += name
        payload = dict()
        payload['devices'] = list()
        response = self.bigip.icr_session.put(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        elif response.status_code == 404:
            return True
        else:
            Log.error('device-group', response.text)
            raise exceptions.ClusterQueryException(response.text)

    @log
    def remove_device(self, name, device_name):
        """Remove device from device group """
        self.remove_devices(name, [device_name])

    @log
    def set_metadata(self, name, cluster_dict):
        """Set metadata on device group """
        if isinstance(cluster_dict, dict):
            str_comment = json.dumps(cluster_dict)
        else:
            str_comment = cluster_dict

        request_url = self.bigip.icr_url + '/cm/device-group/~Common~'
        request_url += name
        payload = dict()
        payload['description'] = base64.encodestring(str_comment)
        response = self.bigip.icr_session.put(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            Log.error('device-group', response.text)
            raise exceptions.ClusterUpdateException(response.text)

    @log
    def get_metadata(self, name):
        """Get metadata on device group """
        request_url = self.bigip.icr_url + '/cm/device-group/~Common~'
        request_url += name + '?$select=name,description'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        str_comment = None
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if response_obj['name'] == name:
                if 'description' in response_obj:
                    str_comment = response_obj['description']
        if str_comment:
            try:
                return json.loads(base64.decodestring(str_comment))
            except Exception:
                try:
                    return base64.decodestring(str_comment)
                except Exception as e:
                    Log.error('device-group', e.message)
                    return str_comment
        return None

    @log
    def remove_metadata(self, name, remove_dict=None):
        """Remove metadata on device group """
        if isinstance(remove_dict, dict):
            existing_metadata = self.get_metadata(name)
            if isinstance(existing_metadata, dict):
                for key in remove_dict:
                    if key in existing_metadata:
                        del(existing_metadata[key])
                return self.set_metadata(name, existing_metadata)
            else:
                return self.set_metadata(name, '')
        else:
            return self.set_metadata(name, '')

    @log
    def update_metadata(self, name, cluster_dict):
        """Update metadata on device group """
        if isinstance(cluster_dict, dict):
            existing_metadata = self.get_metadata(name)
            if isinstance(existing_metadata, dict):
                for key in cluster_dict:
                    existing_metadata[key] = cluster_dict[key]
                return self.set_metadata(name, existing_metadata)
            else:
                return self.set_metadata(name, cluster_dict)
        else:
            return self.set_metadata(name, cluster_dict)

    @log
    def get_traffic_groups(self):
        """Get traffic groups """
        request_url = self.bigip.icr_url + '/cm/traffic-group'
        request_url += '?$select=name'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            traffic_groups = []
            if 'items' in response_obj:
                for tg in response_obj['items']:
                    traffic_groups.append(tg['name'])
            return traffic_groups
        return None

    @log
    def traffic_group_exists(self, name):
        """Does traffic group exist? """
        if name:
            request_url = self.bigip.icr_url + '/cm/traffic-group/'
            request_url += '~Common~' + name
            request_url += '?$select=name'
            response = self.bigip.icr_session.get(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            elif response.status_code == 404:
                return False
            else:
                Log.error('traffic-group', response.text)
                raise exceptions.ClusterQueryException(response.text)

    @log
    def create_traffic_group(self,
                             name=None, autofailback=False,
                             failbacktimer=60, loadfactor=1,
                             floating=True, ha_order=None):
        """Create traffic group """
        request_url = self.bigip.icr_url + '/cm/traffic-group'
        payload = dict()
        payload['name'] = name
        payload['autoFailbackEnabled'] = autofailback
        payload['autoFailbackTime'] = failbacktimer
        payload['haLoadFactor'] = loadfactor
        payload['isFloating'] = floating
        if ha_order:
            ha_order_list = []
            devices = self.bigip.device.get_all_device_names()
            for device in ha_order:
                dev_name = os.path.basename(device)
                if dev_name in devices:
                    ha_order_list.append('/Common/' + dev_name)
            payload['haOrder'] = ha_order_list
        response = self.bigip.icr_session.post(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        elif response.status_code == 409:
            return True
        else:
            Log.error('traffic-group', response.text)
            raise exceptions.ClusterCreationException(response.text)

    @log
    def update_traffic_group(self,
                             name=None, autofailback=False,
                             failbacktimer=60, loadfactor=1,
                             floating=True, ha_order=None):
        """Update traffic group """
        request_url = self.bigip.icr_url + '/cm/traffic-group/'
        request_url += '~Common~' + name
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code >= 400:
            Log.error('traffic-group', response.text)
            raise exceptions.ClusterUpdateException(response.text)
        payload = json.loads(response.text)

        payload['autoFailbackEnabled'] = autofailback
        payload['autoFailbackTime'] = failbacktimer
        payload['haLoadFactor'] = loadfactor
        payload['isFloating'] = floating
        if ha_order:
            ha_order_list = []
            devices = self.bigip.device.get_all_device_names()
            for device in ha_order:
                dev_name = os.path.basename(device)
                if dev_name in devices:
                    ha_order_list.append('/Common/' + dev_name)
            payload['haOrder'] = ha_order_list
        response = self.bigip.icr_session.put(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        elif response.status_code == 409:
            return True
        else:
            Log.error('traffic-group', response.text)
            raise exceptions.ClusterUpdateException(response.text)

    @log
    def delete_traffic_group(self, name):
        """Delete traffic group """
        if name:
            request_url = self.bigip.icr_url + '/cm/traffic-group/'
            request_url += '~Common~' + name
            response = self.bigip.icr_session.delete(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            elif response.status_code == 404:
                return True
            else:
                Log.error('traffic-group', response.text)
                raise exceptions.ClusterDeleteException(response.text)



# Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource


class Monitor(OrganizingCollection):
    def __init__(self, ltm):
        super(Monitor, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [
            HTTPCollection,
            HTTPSCollection,
            DiameterCollection,
            DNSCollection,
            ExternalCollection,
            FirePassCollection,
            FTPCollection,
            GateWay_ICMPCollection,
            ICMPCollection,
            IMAPCollection,
            InBandCollection,
            LDAPCollection,
            Module_ScoreCollection,
            MSSQLCollection,
            MYSQLCollection,
            NNTPCollection,
            NONECollection,
            OracleCollection,
            POP3Collection,
            PostGRESQLCollection,
            RadiusCollection,
            Radius_AccountingCollection,
            Real_ServerCollection,
            RPCCollection,
            SASPCollection,
            ScriptedCollection,
            SIPCollection,
            SMBCollection,
            SMTPCollection,
            SNMP_DCACollection,
            SNMP_DCA_BaseCollection,
            SOAPCollection,
            TCPCollection,
            TCP_EchoCollection,
            TCP_Half_OpenCollection,
            UDPCollection,
            Virtual_LocationCollection,
            WAPCollection,
            WMICollection]


class HTTPCollection(Collection):
    def __init__(self, monitor):
        super(HTTPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [HTTP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:http:httpstate': HTTP}


class UpdateMonitorMixin(object):
    def update(self, **kwargs):
        self.__dict__.pop(u'defaultsFrom', '')
        self._update(**kwargs)


class HTTP(UpdateMonitorMixin, Resource):
    def __init__(self, http_collection):
        super(HTTP, self).__init__(http_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:http:httpstate'


class HTTPSCollection(Collection):
    def __init__(self, monitor):
        super(HTTPSCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [HTTPS]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:https:httpsstate': HTTPS}


class HTTPS(UpdateMonitorMixin, Resource):
    def __init__(self, https_collection):
        super(HTTPS, self).__init__(https_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:https:httpsstate'


class DiameterCollection(Collection):
    def __init__(self, monitor):
        super(DiameterCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Diameter]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:diameter:diameterstate': Diameter}


class Diameter(UpdateMonitorMixin, Resource):
    def __init__(self, diameter_collection):
        super(Diameter, self).__init__(diameter_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:diameter:diameterstate'


class DNSCollection(Collection):
    def __init__(self, monitor):
        super(DNSCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [DNS]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:dns:dnsstate': DNS}


class DNS(UpdateMonitorMixin, Resource):
    def __init__(self, dns_collection):
        super(DNS, self).__init__(dns_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:dns:dnsstate'
        self._meta_data['required_creation_parameters'].update(('qname',))


class ExternalCollection(Collection):
    def __init__(self, monitor):
        super(ExternalCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [External]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:external:externalstate': External}


class External(UpdateMonitorMixin, Resource):
    def __init__(self, external_collection):
        super(External, self).__init__(external_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:external:externalstate'


class FirePassCollection(Collection):
    def __init__(self, monitor):
        super(FirePassCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [FirePass]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:firepass:firepassstate': FirePass}


class FirePass(UpdateMonitorMixin, Resource):
    def __init__(self, firepass_collection):
        super(FirePass, self).__init__(firepass_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:firepass:firepassstate'


class FTPCollection(Collection):
    def __init__(self, monitor):
        super(FTPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [FTP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:ftp:ftpstate': FTP}


class FTP(UpdateMonitorMixin, Resource):
    def __init__(self, ftp_collection):
        super(FTP, self).__init__(ftp_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:ftp:ftpstate'


class GateWay_ICMPCollection(Collection):
    def __init__(self, monitor):
        super(GateWay_ICMPCollection, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [GateWay_ICMP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:gateway-icmp:gateway-icmpstate': GateWay_ICMP}


class GateWay_ICMP(UpdateMonitorMixin, Resource):
    def __init__(self, gateway_icmp_collection):
        super(GateWay_ICMP, self).__init__(gateway_icmp_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:gateway-icmp:gateway-icmpstate'


class ICMPCollection(Collection):
    def __init__(self, monitor):
        super(ICMPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [ICMP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:icmp:icmpstate': ICMP}


class ICMP(UpdateMonitorMixin, Resource):
    def __init__(self, icmp_collection):
        super(ICMP, self).__init__(icmp_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:icmp:icmpstate'


class IMAPCollection(Collection):
    def __init__(self, monitor):
        super(IMAPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [IMAP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:imap:imapstate': IMAP}


class IMAP(UpdateMonitorMixin, Resource):
    def __init__(self, imap_collection):
        super(IMAP, self).__init__(imap_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:imap:imapstate'


class InBandCollection(Collection):
    def __init__(self, monitor):
        super(InBandCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [InBand]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:inband:inbandstate': InBand}


class InBand(UpdateMonitorMixin, Resource):
    def __init__(self, inband_collection):
        super(InBand, self).__init__(inband_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:inband:inbandstate'


class LDAPCollection(Collection):
    def __init__(self, monitor):
        super(LDAPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [LDAP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:ldap:ldapstate': LDAP}


class LDAP(UpdateMonitorMixin, Resource):
    def __init__(self, ldap_collection):
        super(LDAP, self).__init__(ldap_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:ldap:ldapstate'


class Module_ScoreCollection(Collection):
    def __init__(self, monitor):
        super(Module_ScoreCollection, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Module_Score]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:module-score:module-scorestate': Module_Score}


class Module_Score(UpdateMonitorMixin, Resource):
    def __init__(self, gateway_icmp_collection):
        super(Module_Score, self).__init__(gateway_icmp_collection)
        self._meta_data['required_creation_parameters'].update(
            ('snmp-ip-address',))
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:module-score:module-scorestate'


class MYSQLCollection(Collection):
    def __init__(self, monitor):
        super(MYSQLCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [MYSQL]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:mysql:mysqlstate': MYSQL}


class MYSQL(UpdateMonitorMixin, Resource):
    def __init__(self, mysql_collection):
        super(MYSQL, self).__init__(mysql_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:mysql:mysqlstate'


class MSSQLCollection(Collection):
    def __init__(self, monitor):
        super(MSSQLCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [MSSQL]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:mssql:mssqlstate': MSSQL}


class MSSQL(UpdateMonitorMixin, Resource):
    def __init__(self, mssql_collection):
        super(MSSQL, self).__init__(mssql_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:mssql:mssqlstate'


class NNTPCollection(Collection):
    def __init__(self, monitor):
        super(NNTPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [NNTP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:nntp:nntpstate': NNTP}


class NNTP(UpdateMonitorMixin, Resource):
    def __init__(self, nntp_collection):
        super(NNTP, self).__init__(nntp_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:nntp:nntpstate'


class NONECollection(Collection):
    def __init__(self, monitor):
        super(NONECollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [NONE]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:none:nonestate': NONE}


class NONE(UpdateMonitorMixin, Resource):
    def __init__(self, none_collection):
        super(NONE, self).__init__(none_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:none:nonestate'


class OracleCollection(Collection):
    def __init__(self, monitor):
        super(OracleCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Oracle]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:oracle:oraclestate': Oracle}


class Oracle(UpdateMonitorMixin, Resource):
    def __init__(self, oracle_collection):
        super(Oracle, self).__init__(oracle_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:oracle:oraclestate'


class POP3Collection(Collection):
    def __init__(self, monitor):
        super(POP3Collection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [POP3]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:pop3:pop3state': POP3}


class POP3(UpdateMonitorMixin, Resource):
    def __init__(self, pop3_collection):
        super(POP3, self).__init__(pop3_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:pop3:pop3state'


class PostGRESQLCollection(Collection):
    def __init__(self, monitor):
        super(PostGRESQLCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [PostGRESQL]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:postgresql:postgresqlstate': PostGRESQL}


class PostGRESQL(UpdateMonitorMixin, Resource):
    def __init__(self, postgresql_collection):
        super(PostGRESQL, self).__init__(postgresql_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:postgresql:postgresqlstate'


class RadiusCollection(Collection):
    def __init__(self, monitor):
        super(RadiusCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Radius]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:radius:radiusstate': Radius}


class Radius(UpdateMonitorMixin, Resource):
    def __init__(self, radius_collection):
        super(Radius, self).__init__(radius_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:radius:radiusstate'


class Radius_AccountingCollection(Collection):
    def __init__(self, monitor):
        super(Radius_AccountingCollection, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Radius_Accounting]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:radius-accounting:radius-accountingstate':
             Radius_Accounting}


class Radius_Accounting(UpdateMonitorMixin, Resource):
    def __init__(self, radius_accounting_collection):
        super(Radius_Accounting, self).__init__(radius_accounting_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:radius-accounting:radius-accountingstate'


class Real_ServerCollection(Collection):
    def __init__(self, monitor):
        super(Real_ServerCollection, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Real_Server]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:real-server:real-serverstate': Real_Server}


class Real_Server(UpdateMonitorMixin, Resource):
    def __init__(self, real_server_collection):
        super(Real_Server, self).__init__(real_server_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:real-server:real-serverstate'

    def update(self, **kwargs):
        self.__dict__.pop('tmCommand', '')
        self.__dict__.pop('agent', '')
        self.__dict__.pop('method', '')
        super(Real_Server, self).update(**kwargs)


class RPCCollection(Collection):
    def __init__(self, monitor):
        super(RPCCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [RPC]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:rpc:rpcstate': RPC}


class RPC(UpdateMonitorMixin, Resource):
    def __init__(self, rpc_collection):
        super(RPC, self).__init__(rpc_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:rpc:rpcstate'


class SASPCollection(Collection):
    def __init__(self, monitor):
        super(SASPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [SASP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:sasp:saspstate': SASP}


class SASP(UpdateMonitorMixin, Resource):
    def __init__(self, sasp_collection):
        super(SASP, self).__init__(sasp_collection)
        self._meta_data['required_creation_parameters'].update(
            ('primaryAddress',))
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:sasp:saspstate'


class ScriptedCollection(Collection):
    def __init__(self, monitor):
        super(ScriptedCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Scripted]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:scripted:scriptedstate': Scripted}


class Scripted(UpdateMonitorMixin, Resource):
    def __init__(self, scripted_collection):
        super(Scripted, self).__init__(scripted_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:scripted:scriptedstate'


class SIPCollection(Collection):
    def __init__(self, monitor):
        super(SIPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [SIP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:sip:sipstate': SIP}


class SIP(UpdateMonitorMixin, Resource):
    def __init__(self, sip_collection):
        super(SIP, self).__init__(sip_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:sip:sipstate'


class SMBCollection(Collection):
    def __init__(self, monitor):
        super(SMBCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [SMB]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:smb:smbstate': SMB}


class SMB(UpdateMonitorMixin, Resource):
    def __init__(self, smb_collection):
        super(SMB, self).__init__(smb_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:smb:smbstate'


class SMTPCollection(Collection):
    def __init__(self, monitor):
        super(SMTPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [SMTP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:smtp:smtpstate': SMTP}


class SMTP(UpdateMonitorMixin, Resource):
    def __init__(self, smtp_collection):
        super(SMTP, self).__init__(smtp_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:smtp:smtpstate'


class SNMP_DCACollection(Collection):
    def __init__(self, monitor):
        super(SNMP_DCACollection, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [SNMP_DCA]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:snmp-dca:snmp-dcastate': SNMP_DCA}


class SNMP_DCA(UpdateMonitorMixin, Resource):
    def __init__(self, snmp_dca_collection):
        super(SNMP_DCA, self).__init__(snmp_dca_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:snmp-dca:snmp-dcastate'


class SNMP_DCA_BaseCollection(Collection):
    def __init__(self, monitor):
        super(SNMP_DCA_BaseCollection, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [SNMP_DCA_Base]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:snmp-dca-base:snmp-dca-basestate': SNMP_DCA_Base}


class SNMP_DCA_Base(UpdateMonitorMixin, Resource):
    def __init__(self, snmp_dca_base_collection):
        super(SNMP_DCA_Base, self).__init__(snmp_dca_base_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:snmp-dca-base:snmp-dca-basestate'


class SOAPCollection(Collection):
    def __init__(self, monitor):
        super(SOAPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [SOAP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:soap:soapstate': SOAP}


class SOAP(UpdateMonitorMixin, Resource):
    def __init__(self, soap_collection):
        super(SOAP, self).__init__(soap_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:soap:soapstate'


class TCPCollection(Collection):
    def __init__(self, monitor):
        super(TCPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [TCP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:tcp:tcpstate': TCP}


class TCP(UpdateMonitorMixin, Resource):
    def __init__(self, tcp_collection):
        super(TCP, self).__init__(tcp_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:tcp:tcpstate'


class TCP_EchoCollection(Collection):
    def __init__(self, monitor):
        super(TCP_EchoCollection, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [TCP_Echo]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:tcp-echo:tcp-echostate': TCP_Echo}


class TCP_Echo(UpdateMonitorMixin, Resource):
    def __init__(self, tcp_echo_collection):
        super(TCP_Echo, self).__init__(tcp_echo_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:tcp-echo:tcp-echostate'


class TCP_Half_OpenCollection(Collection):
    def __init__(self, monitor):
        super(TCP_Half_OpenCollection, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [TCP_Half_Open]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:tcp-half-open:tcp-half-openstate': TCP_Half_Open}


class TCP_Half_Open(UpdateMonitorMixin, Resource):
    def __init__(self, tcp_half_open_collection):
        super(TCP_Half_Open, self).__init__(tcp_half_open_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:tcp-half-open:tcp-half-openstate'


class UDPCollection(Collection):
    def __init__(self, monitor):
        super(UDPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [UDP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:udp:udpstate': UDP}


class UDP(UpdateMonitorMixin, Resource):
    def __init__(self, udp_collection):
        super(UDP, self).__init__(udp_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:udp:udpstate'


class Virtual_LocationCollection(Collection):
    def __init__(self, monitor):
        super(Virtual_LocationCollection, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Virtual_Location]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:virtual-location:virtual-locationstate':
             Virtual_Location}


class Virtual_Location(UpdateMonitorMixin, Resource):
    def __init__(self, virtual_location_collection):
        super(Virtual_Location, self).__init__(virtual_location_collection)
        self._meta_data['required_creation_parameters'].update(
            ('pool',))
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:virtual-location:virtual-locationstate'


class WAPCollection(Collection):
    def __init__(self, monitor):
        super(WAPCollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [WAP]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:wap:wapstate': WAP}


class WAP(UpdateMonitorMixin, Resource):
    def __init__(self, wap_collection):
        super(WAP, self).__init__(wap_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:wap:wapstate'


class WMICollection(Collection):
    def __init__(self, monitor):
        super(WMICollection, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [WMI]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:wmi:wmistate': WMI}


class WMI(UpdateMonitorMixin, Resource):
    def __init__(self, wmi_collection):
        super(WMI, self).__init__(wmi_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:wmi:wmistate'
        self._meta_data['read_only_attributes'] =\
            ['agent', 'post', 'method']

    def update(self, **kwargs):
        self.__dict__.pop('agent', '')
        self.__dict__.pop('post', '')
        self.__dict__.pop('method', '')
        super(WMI, self).update(**kwargs)



# Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class PolicyCollection(Collection):
    def __init__(self, ltm):
        super(PolicyCollection, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Policy]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:policy:policystate': Policy}


class Policy(Resource):
    def __init__(self, policy_collection):
        super(Policy, self).__init__(policy_collection)
        self._meta_data['required_json_kind'] = 'tm:ltm:policy:policystate'
        self._meta_data['required_creation_parameters'].update(('strategy',))
        temp = {'tm:ltm:policy:rules:rulescollectionstate': RulesCollection}
        self._meta_data['attribute_registry'] = temp


class RulesCollection(Collection):
    def __init__(self, policy):
        super(RulesCollection, self).__init__(policy)
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:policy:rules:rulesstate': Rules}
        self._meta_data['required_json_kind'] =\
            'tm:ltm:policy:rules:rulescollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Rules]


class Rules(Resource):
    def __init__(self, rules_collection):
        super(Rules, self).__init__(rules_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:policy:rules:rulesstate'
        temp = {'tm:ltm:policy:rules:actions:actionscollectionstate':
                ActionsCollection,
                'tm:ltm:policy:rules:conditions:conditionscollectionstate':
                ConditionsCollection}
        self._meta_data['attribute_registry'] = temp


class ActionsCollection(Collection):
    def __init__(self, rules):
        super(ActionsCollection, self).__init__(rules)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:policy:rules:actions:actionscollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Actions]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:policy:rules:actions:actionsstate': Actions}


class Actions(Resource):
    def __init__(self, actions_collection):
        super(Actions, self).__init__(actions_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:policy:rules:actions:actionsstate'


class ConditionsCollection(Collection):
    def __init__(self, rules):
        super(ConditionsCollection, self).__init__(rules)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:policy:rules:conditions:conditionscollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Conditions]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:policy:rules:conditions:conditionsstate': Conditions}


class Conditions(Resource):
    def __init__(self, conditions_collection):
        super(Conditions, self).__init__(conditions_collection)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:policy:rules:conditions:conditionsstate'
